#!/usr/bin/env python

"""Tests for `bunny_storm` package."""

import asyncio
from logging import Logger

import pytest
from aio_pika import IncomingMessage

from tests.conftest import collect_future, consumer_teardown, publisher_teardown
from bunny_storm import AsyncAdapter, RabbitMQConnectionData, Publisher, Consumer


@pytest.fixture(scope="function")
def rpc_executor_config() -> dict:
    return dict(
        receive=dict(
            incoming=dict(
                exchange_name="echo_requests",
                exchange_type="topic",
                routing_key="#",
                queue_name="echo_request_queue",
                durable=True,
                auto_delete=False,
                prefetch_count=10
            )
        )
    )


@pytest.fixture(scope="function")
def rpc_server_config() -> dict:
    return dict(
        receive=dict(
            incoming=dict(
                exchange_name="echo_responses",
                exchange_type="direct",
                routing_key="echo_response",
                queue_name="echo_response_queue",
                durable=True,
                auto_delete=False,
                prefetch_count=1
            )
        ),
        publish=dict(
            outgoing=dict(
                exchange_name="echo_requests",
                exchange_type="topic",
                routing_key="echo_request",
                durable=True,
                auto_delete=False,
                prefetch_count=1
            )
        )
    )


async def async_adapter_teardown(adapter: AsyncAdapter) -> None:
    for consumer in adapter._consumers.values():
        await consumer_teardown(consumer)
    for publisher in adapter._publishers.values():
        await publisher_teardown(publisher)


@pytest.fixture(scope="function")
async def executor_adapter(rabbitmq_connection_data: RabbitMQConnectionData, rpc_executor_config: dict) -> AsyncAdapter:
    adapter = AsyncAdapter(rabbitmq_connection_data, rpc_executor_config)
    yield adapter
    # Teardown
    await async_adapter_teardown(adapter)


@pytest.fixture(scope="function")
async def server_adapter(rabbitmq_connection_data: RabbitMQConnectionData, rpc_server_config: dict) -> AsyncAdapter:
    adapter = AsyncAdapter(rabbitmq_connection_data, rpc_server_config)
    yield adapter
    # Teardown
    await async_adapter_teardown(adapter)


class TestAsyncAdapter:
    _test_future: asyncio.Future = None
    _received_rpc_body: bytes = None

    def test_async_adapter_creation(self, rabbitmq_connection_data: RabbitMQConnectionData) -> None:
        # Arrange
        test_config = dict(
            receive=dict(
                incoming=dict(
                    exchange_name="echo_responses",
                    exchange_type="direct",
                    routing_key="echo_response",
                    queue_name="echo_response_queue",
                    durable=True,
                    auto_delete=False,
                    prefetch_count=1
                ),
                incoming2=dict(
                    exchange_name="echo_responses_2",
                    exchange_type="topic",
                    routing_key="echo_response_2",
                    queue_name="echo_response_queue_2",
                    durable=True,
                    auto_delete=False,
                    prefetch_count=10,
                )
            ),
            publish=dict(
                outgoing=dict(
                    exchange_name="echo_requests",
                    exchange_type="topic",
                    routing_key="echo_request",
                    durable=True,
                    auto_delete=False,
                    prefetch_count=1
                )
            )
        )

        rabbit_adapter = AsyncAdapter(rabbitmq_connection_data=rabbitmq_connection_data, configuration=test_config)

        # Assert
        for publish_config in test_config["publish"].values():
            publisher = rabbit_adapter._publishers[publish_config["exchange_name"]]
            assert isinstance(publisher, Publisher)
            assert publisher._exchange_name == publish_config["exchange_name"]
            assert publisher._exchange_type == publish_config["exchange_type"]
            assert publisher._routing_key == publish_config["routing_key"]
            assert publisher._durable == publish_config["durable"]
            assert publisher._auto_delete == publish_config["auto_delete"]
            assert publisher.channel_config._prefetch_count == publish_config["prefetch_count"]

        for consume_config in test_config["receive"].values():
            consumer = rabbit_adapter._consumers[consume_config["queue_name"]]
            assert isinstance(consumer, Consumer)
            assert consumer._exchange_name == consume_config["exchange_name"]
            assert consumer._exchange_type == consume_config["exchange_type"]
            assert consumer._queue_name == consume_config["queue_name"]
            assert consumer._routing_key == consume_config["routing_key"]
            assert consumer._durable == consume_config["durable"]
            assert consumer._auto_delete == consume_config["auto_delete"]
            assert consumer.channel_config._prefetch_count == consume_config["prefetch_count"]

    @pytest.mark.asyncio
    async def test_async_adapter_consume_publish(self,
                                                 event_loop: asyncio.AbstractEventLoop,
                                                 executor_adapter: AsyncAdapter,
                                                 server_adapter: AsyncAdapter):
        # Arrange
        exchange_name = "echo_requests"
        queue_name = "echo_request_queue"
        expected_message_body = b"test message"

        # Act
        self._test_future = event_loop.create_future()
        await executor_adapter.receive(self.base_message_handler, queue_name)
        await server_adapter.publish(body=expected_message_body, exchange=exchange_name)
        message_body = await collect_future(self._test_future, timeout=10)

        # Assert
        assert executor_adapter.status_check()
        assert server_adapter.status_check()

        assert message_body == expected_message_body

    @pytest.mark.asyncio
    async def test_async_adapter_rpc(self, event_loop, executor_adapter: AsyncAdapter, server_adapter: AsyncAdapter):
        # Arrange
        message_body = b"Double this!"
        expected_result = b"Double this! Double this!"

        # Act
        await executor_adapter.receive(handler=self.rpc_executor_message_handler, queue="echo_request_queue")
        result = await server_adapter.rpc(body=message_body,
                                          receive_queue="echo_response_queue",
                                          publish_exchange="echo_requests",
                                          timeout=10,
                                          ttl=10)

        # Assert
        assert executor_adapter.status_check()
        assert server_adapter.status_check()

        assert self._received_rpc_body == message_body
        assert result == expected_result

    async def base_message_handler(self, logger: Logger, message: IncomingMessage) -> None:
        self._test_future.set_result(message.body)

    async def rpc_executor_message_handler(self, logger: Logger, message: IncomingMessage) -> bytes:
        self._received_rpc_body = message.body
        return message.body + b" " + message.body

    # Error handling tests
    @pytest.mark.asyncio
    async def test_async_adapter_publish_error_handling(self, executor_adapter: AsyncAdapter):
        # Arrange
        bad_message_content = "This should be a bytes object"
        bad_exchange_name = "This exchange doesn't exist"

        # Act/Assert
        with pytest.raises(KeyError):
            await executor_adapter.publish(b"", bad_exchange_name)

        with pytest.raises(Exception):
            await executor_adapter.publish(bad_message_content, "echo_responses")

    @pytest.mark.asyncio
    async def test_async_adapter_receive_error_handling(self, executor_adapter: AsyncAdapter):
        # Arrange
        bad_queue_name = "This queue doesn't exist"

        # Act/Assert
        with pytest.raises(KeyError):
            await executor_adapter.receive(self.base_message_handler, bad_queue_name)

    @pytest.mark.asyncio
    async def test_async_adapter_rpc_bad_queue(self, event_loop, server_adapter: AsyncAdapter):
        # Arrange
        message_body = b"Handle this!"

        # Assert
        with pytest.raises(KeyError):
            await server_adapter.rpc(body=message_body,
                                     receive_queue="Nonexistent queue",
                                     publish_exchange="echo_requests",
                                     timeout=10,
                                     ttl=10)

    @pytest.mark.asyncio
    async def test_async_adapter_rpc_timeout(self,
                                             event_loop: asyncio.AbstractEventLoop,
                                             executor_adapter: AsyncAdapter,
                                             server_adapter: AsyncAdapter):
        # Arrange
        message_body = b"Handle this!"

        # Assert
        await executor_adapter.receive(handler=self.rpc_executor_message_handler, queue="echo_request_queue")
        with pytest.raises(TimeoutError):
            await server_adapter.rpc(body=message_body,
                                     receive_queue="echo_response_queue",
                                     publish_exchange="echo_requests",
                                     timeout=0,
                                     ttl=10)

    @pytest.mark.asyncio
    async def test_stop(self, server_adapter: AsyncAdapter):
        # Act
        await server_adapter.stop(stop_loop=False)

        # Assert
        for consumer in server_adapter.consumers.values():
            assert consumer.channel_config._channel is None

        for publisher in server_adapter.publishers.values():
            assert publisher.channel_config._channel is None

        assert not server_adapter._connection.is_connected()



from bunny_storm import RabbitMQConnectionData


def test_connection_data_creation() -> None:
    # Arrange
    expected_user = "user"
    expected_pass = "pass"
    expected_host = "8.8.8.8"
    default_port = 5672
    default_vhost = "/"

    expected_uri = f"amqp://{expected_user}:{expected_pass}@{expected_host}:{default_port}/"

    # Act
    connection_data = RabbitMQConnectionData(username=expected_user, password=expected_pass, host=expected_host)

    # Assert
    assert connection_data.username == expected_user
    assert connection_data.password == expected_pass
    assert connection_data.host == expected_host
    assert connection_data.port == default_port
    assert connection_data.virtual_host == default_vhost
    assert connection_data.uri() == expected_uri



import asyncio
from logging import Logger
from typing import List, Optional

from aio_pika import RobustChannel, RobustExchange, RobustQueue
from aio_pika.types import CloseCallbackType, Sender

from . import AsyncConnection


class ChannelConfiguration:
    """
    Responsible for the management of a single channel in a given connection.
    Exposes APIs for the declaration of exchanges and queues.
    Automatically recovers from being closed.
    """
    _logger: Logger

    _connection: AsyncConnection
    _loop: asyncio.AbstractEventLoop

    _prefetch_count: int
    _channel_number: int
    _publisher_confirms: bool
    _on_return_raises: bool

    _channel_lock: asyncio.Lock
    _channel: Optional[RobustChannel]
    _started: bool

    _channel_close_callbacks: List[CloseCallbackType]

    def __init__(self, connection: AsyncConnection, logger: Logger, loop: asyncio.AbstractEventLoop = None,
                 prefetch_count: int = 1, channel_number: int = None, publisher_confirms: bool = True,
                 on_return_raises: bool = False):
        """
        :param connection: AsyncConnection instance to create a channel with
        :param logger: Logger
        :param loop: Event loop
        :param prefetch_count: Prefetch count of channel
        :param channel_number: Index of channel (set automatically if None)
        :param publisher_confirms: Whether or not to use publisher confirms
        :param on_return_raises: Whether or not to raise an error in case publishing a message causes a return
        """
        self._logger = logger

        self._connection = connection
        self._loop = loop or asyncio.get_event_loop()

        self._prefetch_count = prefetch_count
        self._channel_number = channel_number
        self._publisher_confirms = publisher_confirms
        self._on_return_raises = on_return_raises

        self._channel_lock = asyncio.Lock()
        self._channel = None
        self._started = False

        self._channel_close_callbacks = []

    @property
    def started(self) -> bool:
        """
        Whether or not the channel has been started
        """
        return self._started

    @property
    def publisher_confirms(self) -> bool:
        return self._publisher_confirms

    @property
    def logger(self) -> Logger:
        return self._logger

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def connection(self) -> AsyncConnection:
        return self._connection

    async def get_default_exchange(self) -> RobustExchange:
        await self.ensure_channel()
        return self._channel.default_exchange

    def add_close_callback(self, callback: CloseCallbackType) -> None:
        """
        Adds a callback which will be called in case the channel is closed. Callbacks are called sequentially.
        :param callback: Callback to add
        """
        self._channel_close_callbacks.append(callback)

    async def ensure_channel(self) -> RobustChannel:
        """
        Ensures that the channel has been started and is open, then returns it.
        If the channel has not been started, starts the channel.
        If the channel is closed, reopens it.
        Uses a Lock to ensure that no race conditions can occur.
        :return: Robust channel held by this instance
        """
        async with self._channel_lock:
            if not self._started:
                await self._start_channel()

            if self._channel.is_closed:
                await self._channel.reopen()

        return self._channel

    async def close_channel(self, exc: BaseException) -> None:
        """
        Resets the channel.
        If the channel is already closed, calls the callback for the channel being closed.
        If the channel is open, closes the channel.
        :param exc: Exception to pass for the channel's closing
        """
        if not self._channel:
            return
        if self._channel.is_closed:
            self._on_channel_close(None, exc)
        else:
            await self._channel.close(exc)

    def _on_channel_close(self, sender: Sender, exc: BaseException) -> None:
        """
        Handles the channel getting closed.
        Prints indicative logs, sets self._started to False, and calls all channel close callbacks.
        :param sender: Closer
        :param exc: Exception which caused the closing
        """
        self.logger.error("Channel closed. Exception info: ")
        self.logger.error(exc, exc_info=True)
        self._started = False
        for callback in self._channel_close_callbacks:
            callback(sender, exc)

    async def _start_channel(self) -> RobustChannel:
        """
        Creates a new channel using the AsyncConnection given to this instance, and sets the relevant parameters.
        :return: Channel created
        """
        self.logger.info("Creating channel")
        connection = await self._connection.get_connection()
        self._channel = await connection.channel(channel_number=self._channel_number,
                                                 publisher_confirms=self._publisher_confirms,
                                                 on_return_raises=self._on_return_raises)
        self._channel.add_close_callback(self._on_channel_close)
        await self._channel.set_qos(prefetch_count=self._prefetch_count)

        self._started = True
        return self._channel

    async def declare_exchange(self,
                               exchange_name: str,
                               exchange_type: str = "direct",
                               durable: bool = None,
                               auto_delete: bool = False,
                               **kwargs) -> RobustExchange:
        """
        Declares an exchange with the given parameters.
        If an exchange with the given name already exists in the channel, returns it instead of creating a new one.
        :param exchange_name: Exchange name
        :param exchange_type: Exchange type
        :param durable: Exchange durability
        :param auto_delete: Whether or not the exchange is auto deleted
        :return: Exchange which was declared or gotten
        """
        await self.ensure_channel()
        self.logger.info(f"Declaring exchange: {exchange_name}")
        if exchange_name not in self._channel._exchanges:
            exchange = await self._channel.declare_exchange(
                name=exchange_name,
                type=exchange_type,
                durable=durable,
                auto_delete=auto_delete,
                **kwargs
            )
            self.logger.info(f"Declared exchange: {exchange_name}")
        else:
            exchange = list(self._channel._exchanges[exchange_name])[0]
        return exchange

    async def declare_queue(self,
                            queue_name: str,
                            exchange: RobustExchange = None,
                            routing_key: str = None,
                            durable: bool = None,
                            exclusive: bool = None,
                            auto_delete: bool = False,
                            **kwargs) -> RobustQueue:
        """
        Declares a queue with the given parameters.
        If a queue with the given name already exists in the channel, gets it instead of creating a new one.
        If an exchange and valid routing key are passed, the queue is bound to the exchange with the routing key.
        :param queue_name: Queue name
        :param exchange: Exchange to bind queue to
        :param routing_key: Routing key to bind queue with
        :param durable: Queue durability
        :param exclusive: Queue exclusivity
        :param auto_delete: Whether or not the queue auto deletes
        :return: Declared or gotten queue
        """
        await self.ensure_channel()
        self.logger.info(f"Declaring queue: {queue_name}")
        if queue_name not in self._channel._queues:
            queue = await self._channel.declare_queue(
                name=queue_name,
                durable=durable,
                exclusive=exclusive,
                auto_delete=auto_delete,
                **kwargs
            )
            self.logger.info(f"Declared queue: {queue_name}")
        else:
            queue = list(self._channel._queues[queue_name])[0]

        if exchange and routing_key:
            self.logger.info(f"Binding queue: {queue_name} to exchange: {exchange}, route: {routing_key}")
            await queue.bind(exchange=exchange, routing_key=routing_key)

        return queue



import asyncio
import logging
import os
import sys
from typing import Union

import pytest

from bunny_storm import RabbitMQConnectionData, AsyncConnection, ChannelConfiguration, Consumer, Publisher, \
    IntentionalCloseChannelError

if sys.platform == 'win32':
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


@pytest.fixture(scope="session")
def rabbitmq_user() -> str:
    return os.getenv("RABBITMQ_USER", "guest")


@pytest.fixture(scope="session")
def rabbitmq_password() -> str:
    return os.getenv("RABBITMQ_PASSWORD", "guest")


@pytest.fixture(scope="session")
def rabbitmq_host() -> str:
    return os.getenv("RABBITMQ_HOST", "localhost")


@pytest.fixture(scope="session")
def rabbitmq_port() -> int:
    return int(os.getenv("RABBITMQ_PORT", "5672"))


@pytest.fixture(scope="session")
def rabbitmq_virtual_host() -> str:
    return os.getenv("RABBITMQ_VIRTUAL_HOST", "vhost")


@pytest.fixture(scope="function")
def rabbitmq_connection_data(rabbitmq_user: str, rabbitmq_password: str, rabbitmq_host: str, rabbitmq_port: int,
                             rabbitmq_virtual_host: str) -> RabbitMQConnectionData:
    connection_data = RabbitMQConnectionData(username=rabbitmq_user,
                                             password=rabbitmq_password,
                                             host=rabbitmq_host,
                                             port=rabbitmq_port,
                                             virtual_host=rabbitmq_virtual_host,
                                             connection_name="test_runner")
    return connection_data


@pytest.fixture(scope="session")
def configuration() -> dict:
    return dict(
        publish=dict(
            exchange_name="test_pub",
            exchange_type="direct",
            routing_key="unit_test",
            durable=False,
            auto_delete=True,
            prefetch_count=1
        ),
        receive=dict(
            exchange_name="test_pub",
            exchange_type="direct",
            routing_key="unit_test",
            queue_name="unit_test",
            durable=False,
            auto_delete=True,
            prefetch_count=1
        )
    )


@pytest.fixture(scope="session")
def logger() -> logging.Logger:
    return logging.getLogger(__name__)


@pytest.fixture(scope="function")
def async_connection(rabbitmq_connection_data: RabbitMQConnectionData, event_loop: asyncio.AbstractEventLoop,
                     logger: logging.Logger) -> AsyncConnection:
    return AsyncConnection(rabbitmq_connection_data, logger, event_loop)


@pytest.fixture(scope="function")
async def channel_config(async_connection: AsyncConnection, logger: logging.Logger,
                         event_loop: asyncio.AbstractEventLoop) -> ChannelConfiguration:
    channel_config = ChannelConfiguration(async_connection, logger, event_loop)
    yield channel_config
    # Teardown
    await channel_configuration_teardown(channel_config)


@pytest.fixture(scope="function")
async def publisher(async_connection: AsyncConnection, logger: logging.Logger, configuration: dict) -> Publisher:
    publisher = Publisher(async_connection, logger, **configuration["publish"])
    yield publisher
    # Teardown
    await publisher_teardown(publisher)


@pytest.fixture(scope="function")
async def consumer(async_connection: AsyncConnection, logger: logging.Logger, configuration: dict) -> Consumer:
    consumer = Consumer(async_connection, logger, **configuration["receive"])
    yield consumer
    # Teardown
    await consumer_teardown(consumer)


async def channel_configuration_teardown(channel_configuration: ChannelConfiguration) -> None:
    if channel_configuration._channel:
        await channel_configuration._channel.close(exc=IntentionalCloseChannelError("Teardown"))
        channel_configuration._channel = None


async def consumer_teardown(consumer: Consumer) -> None:
    if consumer._queue:
        await consumer._queue.delete(if_unused=False, if_empty=False)
        consumer._queue = None
    if consumer._exchange:
        await consumer._exchange.delete(if_unused=False)
        consumer._exchange = None
    await channel_configuration_teardown(consumer.channel_config)


async def publisher_teardown(publisher: Publisher) -> None:
    if publisher._exchange:
        await publisher._exchange.delete(if_unused=False)
        publisher._exchange = None
    await channel_configuration_teardown(publisher.channel_config)


async def collect_future(future: asyncio.Future, timeout: Union[int, float]) -> Union[bytes, None]:
    try:
        await asyncio.wait_for(future, timeout=timeout)
        return future.result()
    except asyncio.exceptions.TimeoutError:
        return



from __future__ import annotations

import asyncio
import functools
from logging import Logger
from typing import Union, Optional

import aio_pika
import aiormq
from aio_pika import Message, RobustExchange

from . import ChannelConfiguration, AsyncConnection, IntentionalCloseChannelError


def _retry_publish_exception(coroutine_function):
    """
    A decorator that wraps a publish function in order to retry it in case of an exception, with dedicated logic
    """

    @functools.wraps(coroutine_function)
    async def publish(self: Publisher,
                      message: Message,
                      routing_key: str = None,
                      mandatory: bool = True,
                      immediate: bool = False,
                      timeout: Union[int, float, None] = None) -> None:
        publish_exception = None

        try:
            await coroutine_function(self, message, routing_key, mandatory, immediate, timeout)
        except aiormq.exceptions.ChannelNotFoundEntity as exc:
            self.logger.error(f"Could not send message, Exchange {self._exchange} was not found, resetting channel")
            publish_exception = exc
        except aio_pika.exceptions.DeliveryError as exc:
            self.logger.error(f"Could not send message, Message {message} was returned, resetting channel")
            publish_exception = exc
        except ValueError as exc:
            self.logger.error("Could not send message, Publisher confirm failed")
            publish_exception = exc
        except (asyncio.exceptions.CancelledError, ConnectionError):
            self.logger.error("Could not send message, Publisher was cancelled")
            publish_exception = self.CLOSE_CONNECTION
        except BaseException as exc:
            self.logger.exception(f"Could not send message, Unknown publisher error: {exc}")
            publish_exception = self.CLOSE_CONNECTION

        if publish_exception:
            if publish_exception == self.CLOSE_CONNECTION:
                # Let robust connection handle the reconnection and wait until it's done
                await self.channel_config.connection.wait_connected()
            else:
                await self.channel_config.close_channel(publish_exception)

            await publish(self, message, routing_key, mandatory, immediate, timeout)

    return publish


class Publisher:
    """
    Responsible for publishing messages to a given exchange with a given routing key
    """
    CLOSE_CONNECTION = "Connection close is needed"

    channel_config: ChannelConfiguration

    _exchange_name: str
    _exchange_type: str
    _routing_key: str
    _durable: bool
    _auto_delete: bool
    _exchange_kwargs: dict
    _exchange: Optional[RobustExchange]

    def __init__(self, connection: AsyncConnection, logger: Logger, exchange_name: str,
                 loop: asyncio.AbstractEventLoop = None, exchange_type: str = "topic", routing_key: str = None,
                 durable: bool = False, auto_delete: bool = False, prefetch_count: int = 1,
                 channel_number: int = None, publisher_confirms: bool = True, on_return_raises: bool = True,
                 exchange_kwargs: dict = None, **kwargs):
        """
        :param connection: AsyncConnection to pass to ChannelConfiguration
        :param logger: Logger
        :param loop: Loop
        :param exchange_name: Exchange name to send messages to
        :param exchange_type: Exchange type
        :param routing_key: Default routing key to send the message with
        :param durable: Exchange durability
        :param auto_delete: Whether or not exchange auto deletes
        :param prefetch_count: Prefetch count for ChannelConfiguration
        :param channel_number: Channel number for ChannelConfiguration
        :param publisher_confirms: Publisher confirms for ChannelConfiguration
        :param on_return_raises: On return raises for ChannelConfiguration
        :param queue_kwargs: Kwargs for queue declaration
        """
        self._channel_config = ChannelConfiguration(
            connection,
            logger,
            loop,
            prefetch_count,
            channel_number,
            publisher_confirms,
            on_return_raises
        )

        self._logger = logger
        self._exchange_name = exchange_name
        self._exchange_type = exchange_type
        self._routing_key = routing_key
        self._durable = durable
        self._auto_delete = auto_delete
        self._exchange_kwargs = exchange_kwargs or dict()
        self._exchange = None

        for key, value in kwargs.items():
            self._logger.warning(f"Publisher received unexpected keyword argument. Key: {key} Value: {value}")

    @property
    def logger(self) -> Logger:
        return self._logger

    @property
    def channel_config(self) -> ChannelConfiguration:
        return self._channel_config

    async def _prepare_publish(self) -> None:
        """
        Ensures the channel is started and declares the desired exchange
        """
        await self.channel_config.ensure_channel()
        self._exchange = await self.channel_config.declare_exchange(
            self._exchange_name,
            exchange_type=self._exchange_type,
            durable=self._durable,
            auto_delete=self._auto_delete,
            **self._exchange_kwargs
        )

    async def _publish_message(self,
                               exchange: RobustExchange,
                               message: Message,
                               routing_key: str,
                               mandatory: bool = True,
                               immediate: bool = False,
                               timeout: Union[int, float, None] = None) -> None:
        self.logger.info(
            f"Publishing message. exchange: {exchange}; routing_key: {routing_key}; message: {message.body[:100]}")
        result = await exchange.publish(
            message=message,
            routing_key=routing_key,
            mandatory=mandatory,
            immediate=immediate,
            timeout=timeout
        )
        if self.channel_config.publisher_confirms and not result:
            raise ValueError("Publisher confirm failed")

    @_retry_publish_exception
    async def publish(self,
                      message: Message,
                      routing_key: str = None,
                      mandatory: bool = True,
                      immediate: bool = False,
                      timeout: Union[int, float, None] = None) -> None:
        """
        Publishes the given message to the desired exchange with the desired routing key
        :param message: Message to publish
        :param routing_key: Routing key to send the message with, defaults to the publisher's
        :param mandatory: Whether or not the message is mandatory
        :param immediate: Whether or not the message should be immediate
        :param timeout: Publish timeout
        """
        await self._prepare_publish()
        await self._publish_message(
            exchange=self._exchange,
            message=message,
            routing_key=routing_key or self._routing_key,
            mandatory=mandatory,
            immediate=immediate,
            timeout=timeout
        )

    @_retry_publish_exception
    async def default_exchange_publish(self,
                                       message: Message,
                                       routing_key: str,
                                       mandatory: bool = True,
                                       immediate: bool = False,
                                       timeout: Union[int, float, None] = None) -> None:
        exchange = await self.channel_config.get_default_exchange()
        await self._publish_message(
            exchange=exchange,
            message=message,
            routing_key=routing_key,
            mandatory=mandatory,
            immediate=immediate,
            timeout=timeout
        )

    async def close(self) -> None:
        """
        Close channel intentionally
        """
        await self.channel_config.close_channel(IntentionalCloseChannelError("Close channel"))



import asyncio

import pytest
from aio_pika import RobustExchange, RobustQueue

from tests.conftest import channel_configuration_teardown
from bunny_storm import AsyncConnection, ChannelConfiguration


@pytest.mark.asyncio
async def test_channel_configuration_creation(async_connection: AsyncConnection, event_loop: asyncio.AbstractEventLoop):
    # Arrange
    channel_config = ChannelConfiguration(
        async_connection,
        async_connection.logger,
        loop=event_loop
    )

    # Act
    channel = await channel_config.ensure_channel()
    channel_copy = await channel_config.ensure_channel()

    # Assert
    assert channel is channel_copy
    assert not channel.is_closed

    # Teardown
    await channel_configuration_teardown(channel_config)


@pytest.mark.asyncio
async def test_channel_configuration_declare_exchange(channel_config: ChannelConfiguration) -> None:
    # Arrange
    expected_exchange_name = "creation_test"
    expected_exchange_type = "direct"
    expected_durable = False
    expected_auto_delete = False

    # Act
    exchange = await channel_config.declare_exchange(expected_exchange_name,
                                                     expected_exchange_type,
                                                     expected_durable,
                                                     expected_auto_delete)
    exchange_copy = await channel_config.declare_exchange(expected_exchange_name,
                                                          expected_exchange_type,
                                                          expected_durable,
                                                          expected_auto_delete)

    # Assert
    assert isinstance(exchange, RobustExchange)
    assert exchange is exchange_copy
    assert exchange.name == expected_exchange_name
    assert exchange._Exchange__type == expected_exchange_type  # Private variable testing is weird
    assert exchange.durable == expected_durable
    assert exchange.auto_delete == expected_auto_delete


@pytest.mark.asyncio
async def test_channel_configuration_declare_exchange_queue(channel_config: ChannelConfiguration) -> None:
    # Arrange
    expected_exchange_name = "creation_test"
    expected_exchange_type = "direct"
    expected_queue_name = "creation_test_q"
    expected_routing_key = "#"
    expected_durable = False
    expected_exclusive = True
    expected_auto_delete = False

    # Act
    exchange = await channel_config.declare_exchange(exchange_name=expected_exchange_name,
                                                     exchange_type=expected_exchange_type,
                                                     durable=expected_durable,
                                                     auto_delete=expected_auto_delete)
    queue = await channel_config.declare_queue(expected_queue_name,
                                               exchange=exchange,
                                               routing_key=expected_routing_key,
                                               durable=expected_durable,
                                               exclusive=expected_exclusive,
                                               auto_delete=expected_auto_delete)
    queue_copy = await channel_config.declare_queue(expected_queue_name,
                                                    exchange=exchange,
                                                    routing_key=expected_routing_key,
                                                    durable=expected_durable,
                                                    exclusive=expected_exclusive,
                                                    auto_delete=expected_auto_delete)

    # Assert
    assert isinstance(queue, RobustQueue)
    assert queue is queue_copy
    assert queue.durable == expected_durable
    assert queue.exclusive == expected_exclusive
    assert queue.auto_delete == expected_auto_delete
    assert list(queue._bindings) == [(exchange, expected_routing_key)]


@pytest.mark.asyncio
async def test_channel_configuration_close_reopen(channel_config: ChannelConfiguration):
    # Act
    await channel_config.ensure_channel()
    await channel_config._channel.close()

    # Assert
    assert channel_config._channel.is_closed

    # Act
    await channel_config.ensure_channel()

    # Assert
    assert not channel_config._channel.is_closed



from dataclasses import dataclass


@dataclass
class RabbitMQConnectionData:
    """
    Dataclass responsible for organizing RabbitMQ connection credentials and parameters.
    Allows to name connections using connection_name.
    Use uri() to get the connection URI which corresponds to the credentials and parameters given.
    """
    username: str = "guest"
    password: str = "guest"
    host: str = "localhost"
    port: int = 5672
    virtual_host: str = "/"
    connection_name: str = ""

    def uri(self) -> str:
        """
        Creates connection URI for a RabbitMQ server with the given connection credentials.
        :return: Connection URI
        """
        vhost = "" if self.virtual_host == "/" else self.virtual_host
        name_query = f"?name={self.connection_name}" if self.connection_name else ""
        return f"amqp://{self.username}:{self.password}@{self.host}:{self.port}/{vhost}{name_query}"



import asyncio
import functools
from logging import Logger
from types import FunctionType
from typing import Tuple, Optional

import aiormq
from aio_pika.queue import ConsumerTag
from aio_pika.types import Sender

from . import ChannelConfiguration, IntentionalCloseChannelError, AsyncConnection


class Consumer:
    """
    Responsible for consuming messages from a given queue (or queue and exchange) and running the given message handler
    """
    _channel_config: ChannelConfiguration

    _exchange_name: str
    _exchange_type: str
    _queue_name: str
    _routing_key: str
    _durable: bool
    _exclusive_queue: bool
    _auto_delete: bool
    _exchange_kwargs: dict
    _queue_kwargs: dict

    _should_consume: bool
    _consume_params: Optional[Tuple[FunctionType, FunctionType, bool]]
    _consumer_tag: Optional[ConsumerTag]
    _consumer_lock: asyncio.Lock

    def __init__(self, connection: AsyncConnection, logger: Logger, loop: asyncio.AbstractEventLoop = None,
                 exchange_name: str = None, exchange_type: str = "topic", queue_name: str = "", routing_key: str = None,
                 durable: bool = False, exclusive_queue: bool = False, auto_delete: bool = False,
                 prefetch_count: int = 1, channel_number: int = None, publisher_confirms: bool = True,
                 on_return_raises: bool = False, exchange_kwargs: dict = None, queue_kwargs: dict = None, **kwargs):
        """
        :param connection: AsyncConnection to pass to ChannelConfiguration
        :param logger: Logger
        :param loop: Loop
        :param exchange_name: Exchange name to bind queue to
        :param exchange_type: Exchange type
        :param queue_name: Queue name
        :param routing_key: Routing key for queue in exchange
        :param durable: Queue/exchange durability
        :param exclusive_queue: Queue exclusivity
        :param auto_delete: Whether or not queue/exchange auto delete
        :param prefetch_count: Prefetch count for ChannelConfiguration
        :param channel_number: Channel number for ChannelConfiguration
        :param publisher_confirms: Publisher confirms for ChannelConfiguration
        :param on_return_raises: On return raises for ChannelConfiguration
        :param exchange_kwargs: Kwargs for exchange declaration
        :param queue_kwargs: Kwargs for queue declaration
        """
        self._channel_config = ChannelConfiguration(
            connection,
            logger,
            loop,
            prefetch_count,
            channel_number,
            publisher_confirms,
            on_return_raises
        )
        self._channel_config.add_close_callback(self._on_channel_close)

        self._loop = self._channel_config.loop
        self._logger = logger
        self._exchange_name = exchange_name
        self._exchange_type = exchange_type
        self._queue_name = queue_name
        self._routing_key = routing_key
        self._durable = durable
        self._exclusive_queue = exclusive_queue
        self._auto_delete = auto_delete
        self._exchange_kwargs = exchange_kwargs or dict()
        self._queue_kwargs = queue_kwargs or dict()

        self._exchange = None
        self._queue = None
        self._should_consume = False
        self._consume_params = None
        self._consumer_tag = None
        self._consumer_lock = asyncio.Lock()

        for key, value in kwargs.items():
            self._logger.warning(f"Consumer received unexpected keyword argument. Key: {key} Value: {value}")

    @property
    def logger(self) -> Logger:
        return self._logger

    @property
    def channel_config(self) -> ChannelConfiguration:
        return self._channel_config

    @property
    def consumer_tag(self) -> ConsumerTag:
        return self._consumer_tag

    def _on_channel_close(self, sender: Sender, exc: BaseException) -> None:
        """
        Channel close callback which checks if we want to resume message consumption after the channel has been closed.
        :param sender: Closer
        :param exc: Close exception
        """
        if self._should_consume and not isinstance(exc, IntentionalCloseChannelError):
            self._loop.create_task(self.consume(*self._consume_params))

    async def _prepare_consume(self) -> None:
        """
        Prepares channel, queue, and exchange (if necessary) so the instance can begin consuming messages.
        """
        await self.channel_config.ensure_channel()
        if self._exchange_name:
            self._exchange = await self.channel_config.declare_exchange(
                self._exchange_name,
                exchange_type=self._exchange_type,
                durable=self._durable,
                auto_delete=self._auto_delete,
                **self._exchange_kwargs
            )
        else:
            self._exchange = None
        self._queue = await self.channel_config.declare_queue(
            queue_name=self._queue_name,
            exchange=self._exchange,
            routing_key=self._routing_key,
            durable=self._durable,
            exclusive=self._exclusive_queue,
            auto_delete=self._auto_delete,
            **self._queue_kwargs
        )

    async def consume(self, on_message_callback, handler=None, no_ack: bool = False) -> None:
        """
        Begins consuming messages and triggering the given callback for each message consumed.
        :param on_message_callback: Callback to consume with
        :param handler: Handler to pass on_message_callback
        :param no_ack: Whether or not we want to skip ACKing the messages
        """
        self.logger.info(f"[start consuming] routing key: {self._routing_key}; queue name: {self._queue_name}")
        await self._prepare_consume()
        self._should_consume = True
        self._consume_params = (on_message_callback, handler, no_ack)
        callback = on_message_callback if handler is None else functools.partial(on_message_callback, handler=handler)

        try:
            async with self._consumer_lock:
                self._consumer_tag = await self._queue.consume(callback=callback, no_ack=no_ack,
                                                               consumer_tag=self._consumer_tag)
        except aiormq.exceptions.ChannelNotFoundEntity as exc:
            self.logger.error(f"Queue {self._queue} was not found, resetting channel")
            self._on_channel_close(None, exc)
        except aiormq.exceptions.DuplicateConsumerTag:
            # In case this consumer is already consuming the given queue, do not call consume again. Otherwise a
            # consumer leak will occur.
            self.logger.warning(f"Attempted to consume already consumed queue {self._queue} with tag "
                                f"`{self._consumer_tag}")

    async def close(self) -> None:
        """
        Close channel intentionally
        """
        await self.channel_config.close_channel(IntentionalCloseChannelError("Close channel"))
